<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 08/31/2003 *
 * ===================                          Last change: 11/23/2004 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : admin-inc.php                                    *
 * -------------------------------------------------------------------- *
 * Short description : Administrative related functions                 *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Fuer die Administration benoetigte Funktionen    *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009 - 2011 by Mailer Developer Team                   *
 * For more information visit: http://www.mxchange.org                  *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// Register an administrator account
function addAdminAccount ($adminLogin, $passHash, $adminEmail) {
	// Login does already exist
	$ret = 'already';

	// Lookup the admin
	$result = SQL_QUERY_ESC("SELECT `id` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `login`='%s' LIMIT 1",
		array($adminLogin), __FUNCTION__, __LINE__);

	// Is the entry there?
	if (SQL_HASZERONUMS($result)) {
		// Ok, let's create the admin login
		SQL_QUERY_ESC("INSERT INTO `{?_MYSQL_PREFIX?}_admins` (`login`, `password`, `email`) VALUES ('%s', '%s', '%s')",
			array(
				$adminLogin,
				$passHash,
				$adminEmail
			), __FUNCTION__, __LINE__);

		// All done
		$ret = 'done';
	} // END - if

	// Free memory
	SQL_FREERESULT($result);

	// Return result
	return $ret;
}

// This function will be executed when the admin is not logged in and has submitted his login data
function ifAdminLoginDataIsValid ($adminLogin, $adminPassword) {
	// First of all, no admin login is found, so the admin hash is null
	$ret = '404';
	$adminHash = null;

	// Get admin id from login
	$adminId = getAdminId($adminLogin);

	// Continue only with found admin ids
	if ($adminId > 0) {
		// Then we need to lookup the login name by getting the admin hash
		$adminHash = getAdminHash($adminId);

		// If this is fine, we can continue
		if ($adminHash != '-1') {
			// Get admin id and set it as current
			setCurrentAdminId($adminId);

			// Now, we need to encode the password in the same way the one is encoded in database
			$testHash = generateHash($adminPassword, $adminHash);

			// If they both match, the login data is valid
			if ($testHash == $adminHash) {
				// All fine
				$ret = 'done';
			} else {
				// Did not match!
				$ret = 'password';
			}
		} // END - if
	} // END - if

	// Prepare data array
	$data = array(
		'id'         => $adminId,
		'login'      => $adminLogin,
		'plain_pass' => $adminPassword,
		'pass_hash'  => $adminHash
	);

	// Run a special filter
	runFilterChain('do_admin_login_' . $ret, $data);

	// Return status
	return $ret;
}

// Only be executed on cookie checking
function ifAdminCookiesAreValid ($adminLogin, $passHash) {
	// First of all, no admin login is found
	$ret = '404';

	// Then we need to lookup the login name by getting the admin hash
	$adminHash = getAdminHash($adminLogin);

	// If this is fine, we can continue
	if ($adminHash != '-1') {
		// Now, we need to encode the password in the same way the one is encoded in database
		$testHash = encodeHashForCookie($adminHash);
		//* DEBUG: */ debugOutput('adminLogin=' . $adminLogin . ',passHash='.$passHash.',adminHash='.$adminHash.',testHash='.$testHash);

		// If they both match, the login data is valid
		if ($testHash == $passHash) {
			// All fine
			$ret = 'done';
		} else {
			// Set status
			$ret = 'password';
		}
	} // END - if

	// Return status
	//* DEBUG: */ debugOutput('ret='.$ret);
	return $ret;
}

// Do an admin action
function doAdminAction () {
	// Get default what
	$what = getWhat();

	//* DEBUG: */ debugOutput(__LINE__.'*'.$what.'/'.getModule().'/'.getAction().'/'.getWhat().'*');

	// Remove any spaces from variable
	if (empty($what)) {
		// Default admin action is the overview page
		$what = 'overview';
	} else {
		// Secure it
		$what = secureString($what);
	}

	// Get action value
	$action = getActionFromModuleWhat(getModule(), $what);

	// Load welcome template
	if (isExtensionActive('admins')) {
		// @TODO This and the next getCurrentAdminId() call might be moved into the templates?
		$content['welcome'] = loadTemplate('admin_welcome_admins', true, getCurrentAdminId());
	} else {
		$content['welcome'] = loadTemplate('admin_welcome', true, getCurrentAdminId());
	}

	// Load header, footer, render menu
	$content['header'] = loadTemplate('admin_header' , true, $content);
	$content['footer'] = loadTemplate('admin_footer' , true, $content);
	$content['menu']   = addAdminMenu($action, $what, true);

	// Tableset header
	loadTemplate('admin_main_header', false, $content);

	// Check if action/what pair is valid
	$result_action = SQL_QUERY_ESC("SELECT
	`id`
FROM
	`{?_MYSQL_PREFIX?}_admin_menu`
WHERE
	`action`='%s' AND
	(
		(
			`what`='%s' AND `what` != 'overview'
		) OR (
			(
				`what`='' OR `what` IS NULL
			) AND (
				'%s'='overview'
			)
		)
	)
LIMIT 1",
		array(
			$action,
			$what,
			$what
		), __FUNCTION__, __LINE__);

	// Do we have an entry?
	if (SQL_NUMROWS($result_action) == 1) {
		// Is valid but does the inlcude file exists?
		$inc = sprintf("inc/modules/admin/action-%s.php", $action);
		if ((isIncludeReadable($inc)) && (isMenuActionValid('admin', $action, $what)) && ($GLOBALS['acl_allow'] === true)) {
			// Ok, we finally load the admin action module
			loadInclude($inc);
		} elseif ($GLOBALS['acl_allow'] === false) {
			// Access denied
			loadTemplate('admin_menu_failed', false, '{%message,ADMIN_ACCESS_DENIED=' . $what . '%}');
		} else {
			// Include file not found :-(
			loadTemplate('admin_menu_failed', false, '{%message,ADMIN_ACTION_404=' . $action . '%}');
		}
	} else {
		// Invalid action/what pair found
		loadTemplate('admin_menu_failed', false, '{%message,ADMIN_ACTION_INVALID=' . $action . '/' . $what . '%}');
	}

	// Free memory
	SQL_FREERESULT($result_action);

	// Tableset footer
	loadTemplate('admin_main_footer', false, $content);
}

// Checks wether current admin is allowed to access given action/what combination
// (only one is allowed to be null!)
function isAdminAllowedAccessMenu ($action, $what = null) {
	// Do we have cache?
	if (!isset($GLOBALS[__FUNCTION__][$action][$what])) {
		// ACL is always 'allow' when no ext-admins is installed
		// @TODO This can be rewritten into a filter
		$GLOBALS[__FUNCTION__][$action][$what] = ((!isExtensionInstalledAndNewer('admins', '0.2.0')) || (isAdminsAllowedByAcl($action, $what)));
	} // END - if

	// Return the cached value
	return $GLOBALS[__FUNCTION__][$action][$what];
}

// Adds an admin menu
function addAdminMenu ($action, $what, $return = false) {
	// Init variables
	$SUB = false;
	$OUT = '';

	// Menu descriptions
	$GLOBALS['menu']['description'] = array();
	$GLOBALS['menu']['title'] = array();

	// Build main menu
	$result_main = SQL_QUERY("SELECT
	`action`, `title`, `descr`
FROM
	`{?_MYSQL_PREFIX?}_admin_menu`
WHERE
	(`what`='' OR `what` IS NULL)
ORDER BY
	`sort` ASC,
	`id` DESC", __FUNCTION__, __LINE__);

	// Do we have entries?
	if (!SQL_HASZERONUMS($result_main)) {
		$OUT .= '<ul class="admin_menu_main">';
		// @TODO Rewrite this to $content = SQL_FETCHARRAY()
		while (list($menu, $title, $descr) = SQL_FETCHROW($result_main)) {
			// Filename
			$inc = sprintf("inc/modules/admin/action-%s.php", $menu);

			// Is the file readable?
			$readable = isIncludeReadable($inc);

			// Is the current admin allowed to access this 'action' menu?
			if (isAdminAllowedAccessMenu($menu)) {
				if ($SUB === false) {
					// Insert compiled menu title and description
					$GLOBALS['menu']['title'][$menu]      = $title;
					$GLOBALS['menu']['description'][$menu] = $descr;
				} // END - if
				$OUT .= '<li class="admin_menu">
<div class="nobr"><strong>&middot;</strong>&nbsp;';

				if ($readable === true) {
					if (($menu == $action) && (empty($what))) {
						$OUT .= '<strong>';
					} else {
						$OUT .= '[<a href="{%url=modules.php?module=admin&amp;action=' . $menu . '%}">';
					}
				} else {
					$OUT .= '<em style="cursor:help" class="notice" title="{%message,ADMIN_MENU_ACTION_404_TITLE=' . $menu . '%}">';
				}

				$OUT .= $title;

				if ($readable === true) {
					if (($menu == $action) && (empty($what))) {
						$OUT .= '</strong>';
					} else {
						$OUT .= '</a>]';
					}
				} else {
					$OUT .= '</em>';
				}

				$OUT .= '</div>
</li>';

				// Check for menu entries
				$result_what = SQL_QUERY_ESC("SELECT
	`what`, `title`, `descr`
FROM
	`{?_MYSQL_PREFIX?}_admin_menu`
WHERE
	`action`='%s' AND
	`what` != '' AND
	`what` IS NOT NULL
ORDER BY
	`sort` ASC,
	`id` DESC",
					array($menu), __FUNCTION__, __LINE__);

				// Remember the count for later checks
				setAdminMenuHasEntries($menu, ((!SQL_HASZERONUMS($result_what)) && ($action == $menu)));

				// Do we have entries?
				if ((ifAdminMenuHasEntries($menu)) && (!SQL_HASZERONUMS($result_what))) {
					$GLOBALS['menu']['description'] = array();
					$GLOBALS['menu']['title'] = array();
					$SUB = true;
					$OUT .= '<li class="admin_menu_sub"><ul class="admin_menu_sub">';
					// @TODO Rewrite this to $content = SQL_FETCHARRAY()
					while (list($what_sub, $title_what, $desc_what) = SQL_FETCHROW($result_what)) {
						// Filename
						$inc = sprintf("inc/modules/admin/what-%s.php", $what_sub);

						// Is the file readable?
						$readable = isIncludeReadable($inc);

						// Is the current admin allowed to access this 'what' menu?
						if (isAdminAllowedAccessMenu(null, $what_sub)) {
							// Insert compiled title and description
							$GLOBALS['menu']['title'][$what_sub]      = $title_what;
							$GLOBALS['menu']['description'][$what_sub] = $desc_what;
							$OUT .= '<li class="admin_menu">
<div class="nobr"><strong>--&gt;</strong>&nbsp;';
							if ($readable === true) {
								if ($what == $what_sub) {
									$OUT .= '<strong>';
								} else {
									$OUT .= '[<a href="{%url=modules.php?module=admin&amp;what=' . $what_sub . '%}">';
								}
							} else {
								$OUT .= '<em style="cursor:help" class="notice" title="{%message,ADMIN_MENU_WHAT_404_TITLE=' . $what_sub . '%}">';
							}

							$OUT .= $title_what;

							if ($readable === true) {
								if ($what == $what_sub) {
									$OUT .= '</strong>';
								} else {
									$OUT .= '</a>]';
								}
							} else {
								$OUT .= '</em>';
							}
							$OUT .= '</div>
</li>';
						} // END - if
					} // END - while

					// Free memory
					SQL_FREERESULT($result_what);
					$OUT .= '</ul>
</li>';
				} // END - if
			} // END - if
		} // END - while

		// Free memory
		SQL_FREERESULT($result_main);
		$OUT .= '</ul>';
	}

	// Is there a cache instance again?
	// Return or output content?
	if ($return === true) {
		return $OUT;
	} else {
		outputHtml($OUT);
	}
}

// Create member selection box
function addMemberSelectionBox ($def = 0, $add_all = false, $return = false, $none = false, $field = 'userid') {
	// Output selection form with all confirmed user accounts listed
	$result = SQL_QUERY("SELECT `userid`, `surname`, `family` FROM `{?_MYSQL_PREFIX?}_user_data` ORDER BY `userid` ASC", __FUNCTION__, __LINE__);

	// Default output
	$OUT = '';

	// USe this only for adding points (e.g. adding refs really makes no sence ;-) )
	if ($add_all === true)   $OUT = '      <option value="all">{--ALL_MEMBERS--}</option>';
	 elseif ($none === true) $OUT = '      <option value="0">{--SELECT_NONE--}</option>';

	while ($content = SQL_FETCHARRAY($result)) {
		$OUT .= '<option value="' . bigintval($content['userid']) . '"';
		if ($def == $content['userid']) $OUT .= ' selected="selected"';
		$OUT .= '>' . $content['surname'] . ' ' . $content['family'] . ' (' . bigintval($content['userid']) . ')</option>';
	} // END - while

	// Free memory
	SQL_FREERESULT($result);

	if ($return === false) {
		// Remeber options in constant
		$content['form_selection'] = $OUT;
		$content['what']             = getWhat();

		// Load template
		loadTemplate('admin_form_selection_box', false, $content);
	} else {
		// Return content in selection frame
		return '<select class="form_select" name="' . handleFieldWithBraces($field) . '" size="1">' . $OUT . '</select>';
	}
}

// Create a menu selection box for given menu system
// @TODO Try to rewrite this to adminAddMenuSelectionBox()
// @DEPRECATED
function adminMenuSelectionBox_DEPRECATED ($mode, $default = '', $defid = '') {
	$what = "`what` != '' AND `what` IS NOT NULL";
	if ($mode == 'action') $what = "(`what`='' OR `what` IS NULL) AND `action` != 'login'";

	$result = SQL_QUERY_ESC("SELECT `%s` AS `menu`, `title` FROM `{?_MYSQL_PREFIX?}_admin_menu` WHERE ".$what." ORDER BY `sort` ASC",
		array($mode), __FUNCTION__, __LINE__);
	if (!SQL_HASZERONUMS($result)) {
		// Load menu as selection
		$OUT = '<select name="' . $mode . '_menu';
		if ((!empty($defid)) || ($defid == '0')) $OUT .= '[' . $defid . ']';
		$OUT .= '" size="1" class="form_select">
	<option value="">{--SELECT_NONE--}</option>';
		// Load all entries
		while ($content = SQL_FETCHARRAY($result)) {
			$OUT .= '<option value="' . $content['menu'] . '"';
			if ((!empty($default)) && ($default == $content['menu'])) $OUT .= ' selected="selected"';
			$OUT .= '>' . $content['title'] . '</option>';
		} // END - while

		// Free memory
		SQL_FREERESULT($result);

		// Add closing select-tag
		$OUT .= '</select>';
	} else {
		// No menus???
		$OUT = '{--ADMIN_PROBLEM_NO_MENU--}';
	}

	// Return output
	return $OUT;
}

// Wrapper for $_POST and adminSaveSettings
function adminSaveSettingsFromPostData ($tableName = '_config', $whereStatement = '`config`=0', $translateComma = array(), $alwaysAdd = false, $displayMessage = true) {
	// Get the array
	$postData = postRequestArray();

	// Call the lower function
	adminSaveSettings($postData, $tableName, $whereStatement, $translateComma, $alwaysAdd, $displayMessage);
}

// Save settings to the database
function adminSaveSettings (&$postData, $tableName = '_config', $whereStatement = '`config`=0', $translateComma = array(), $alwaysAdd = false, $displayMessage = true) {
	// Prepare all arrays, variables
	$tableData = array();
	$skip = false;

	// Now, walk through all entries and prepare them for saving
	foreach ($postData as $id => $val) {
		// Process only formular field but not submit buttons ;)
		if ($id != 'ok') {
			// Do not save the ok value
			convertSelectionsToEpocheTime($postData, $tableData, $id, $skip);

			// Shall we process this id? It muss not be empty, of course
			if (($skip === false) && (!empty($id)) && ((!isset($GLOBALS['skip_config'][$id]))) || ($tableName != '_config')) {
				// Translate the value? (comma to dot!)
				if ((is_array($translateComma)) && (in_array($id, $translateComma))) {
					// Then do it here... :)
					$val = convertCommaToDot($val);
				} // END - if

				// Shall we add numbers or strings?
				$test = (float) $val;
				if ('' . $val . '' == '' . $test . '') {
					// Add numbers
					$tableData[] = sprintf("`%s`=%s", $id, $test);
				} elseif (is_null($val)) {
					// Add NULL
					$tableData[] = sprintf("`%s`=NULL", $id);
				} else {
					// Add strings
					$tableData[] = sprintf("`%s`='%s'", $id, trim($val));
				}

				// Do not add a config entry twice
				$GLOBALS['skip_config'][$id] = true;

				// Update current configuration
				setConfigEntry($id, $val);
			} // END - if
		} // END - if
	} // END - foreach

	// Check if entry does exist
	$result = false;
	if ($alwaysAdd === false) {
		if (!empty($whereStatement)) {
			$result = SQL_QUERY("SELECT * FROM `{?_MYSQL_PREFIX?}" . $tableName . "` WHERE " . $whereStatement . " LIMIT 1", __FUNCTION__, __LINE__);
		} else {
			$result = SQL_QUERY("SELECT * FROM `{?_MYSQL_PREFIX?}" . $tableName . "` LIMIT 1", __FUNCTION__, __LINE__);
		}
	} // END - if

	if (SQL_NUMROWS($result) == 1) {
		// "Implode" all data to single string
		$updatedData = implode(', ', $tableData);

		// Generate SQL string
		$sql = sprintf("UPDATE `{?_MYSQL_PREFIX?}%s` SET %s WHERE %s LIMIT 1",
			$tableName,
			$updatedData,
			$whereStatement
		);
	} else {
		// Add Line (does only work with AUTO_INCREMENT!
		$keys = array(); $values = array();
		foreach ($tableData as $entry) {
			// Split up
			$line = explode('=', $entry);
			$keys[] = $line[0];
			$values[] = $line[1];
		} // END - foreach

		// Add both in one line
		$keys = implode('`, `', $keys);
		$values = implode(', ', $values);

		// Generate SQL string
		$sql = sprintf("INSERT INTO `{?_MYSQL_PREFIX?}%s` (%s) VALUES (%s)",
			$tableName,
			$keys,
			$values
		);
	}

	// Free memory
	SQL_FREERESULT($result);

	// Simply run generated SQL string
	SQL_QUERY($sql, __FUNCTION__, __LINE__);

	// Remember affected rows
	$affected = SQL_AFFECTEDROWS();

	// Rebuild cache
	rebuildCache('config', 'config');

	// Settings saved, so display message?
	if ($displayMessage === true) displayMessage('{--SETTINGS_SAVED--}');

	// Return affected rows
	return $affected;
}

// Generate a selection box
function adminAddMenuSelectionBox ($menu, $type, $name, $default = '') {
	// Open the requested menu directory
	$menuArray = getArrayFromDirectory(sprintf("inc/modules/%s/", $menu), $type . '-', false, false);

	// Init the selection box
	$OUT = '<select name="' . $name . '" class="form_select" size="1"><option value="">{--ADMIN_IS_TOP_MENU--}</option>';

	// Walk through all files
	foreach ($menuArray as $file) {
		// Is this a PHP script?
		if ((!isDirectory($file)) && (strpos($file, '' . $type . '-') > -1) && (strpos($file, '.php') > 0)) {
			// Then test if the file is readable
			$test = sprintf("inc/modules/%s/%s", $menu, $file);

			// Is the file there?
			if (isIncludeReadable($test)) {
				// Extract the value for what=xxx
				$part = substr($file, (strlen($type) + 1));
				$part = substr($part, 0, -4);

				// Is that part different from the overview?
				if ($part != 'overview') {
					$OUT .= '<option value="' . $part . '"';
					if ($part == $default) $OUT .= ' selected="selected"';
					$OUT .= '>' . $part . '</option>';
				} // END - if
			} // END - if
		} // END - if
	} // END - foreach

	// Close selection box
	$OUT .= '</select>';

	// Return contents
	return $OUT;
}

// Creates a user-profile link for the admin. This function can also be used for many other purposes
function generateUserProfileLink ($userid, $title = '', $what = 'list_user') {
	if (($title == '') && (isValidUserId($userid))) {
		// Set userid as title
		$title = $userid;
	} elseif (!isValidUserId($userid)) {
		// User id zero is invalid
		return '<strong>' . makeNullToZero($userid) . '</strong>';
	}

	if (($title == '0') && ($what == 'list_refs')) {
		// Return title again
		return $title;
	} elseif (isExtensionActive('nickname')) {
		// Get nickname
		$nick = getNickname($userid);

		// Is it not empty, use it as title else the userid
		if (!empty($nick)) {
			$title = $nick . '(' . $userid . ')';
		} else {
			$title = $userid;
		}
	}

	// Return link
	return '[<a href="{%url=modules.php?module=admin&amp;what=' . $what . '&amp;userid=' . $userid . '%}" title="{--ADMIN_USER_PROFILE_TITLE--}">' . $title . '</a>]';
}

// Check "logical-area-mode"
function adminGetMenuMode () {
	// Set the default menu mode as the mode for all admins
	$mode = 'global';

	// If sql_patches is up-to-date enough, use the configuration
	if (isExtensionInstalledAndNewer('sql_patches', '0.3.2')) {
		$mode = getAdminMenu();
	} // END - if

	// Backup it
	$adminMode = $mode;

	// Get admin id
	$adminId = getCurrentAdminId();

	// Check individual settings of current admin
	if (isset($GLOBALS['cache_array']['admin']['la_mode'][$adminId])) {
		// Load from cache
		$adminMode = $GLOBALS['cache_array']['admin']['la_mode'][$adminId];
		incrementStatsEntry('cache_hits');
	} elseif (isExtensionInstalledAndNewer('admins', '0.6.7')) {
		// Load from database when version of 'admins' is enough
		$result = SQL_QUERY_ESC("SELECT `la_mode` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `id`=%s LIMIT 1",
			array($adminId), __FUNCTION__, __LINE__);

		// Do we have an entry?
		if (SQL_NUMROWS($result) == 1) {
			// Load data
			list($adminMode) = SQL_FETCHROW($result);
		} // END - if

		// Free memory
		SQL_FREERESULT($result);
	}

	// Check what the admin wants and set it when it's not the default mode
	if ($adminMode != 'global') {
		$mode = $adminMode;
	} // END - if

	// Return admin-menu's mode
	return $mode;
}

// Change activation status
function adminChangeActivationStatus ($IDs, $table, $row, $idRow = 'id') {
	$count = '0';
	if ((is_array($IDs)) && (count($IDs) > 0)) {
		// "Walk" all through and count them
		foreach ($IDs as $id => $selected) {
			// Secure the id number
			$id = bigintval($id);

			// Should always be set... ;-)
			if (!empty($selected)) {
				// Determine new status
				$result = SQL_QUERY_ESC("SELECT %s FROM `{?_MYSQL_PREFIX?}_%s` WHERE %s=%s LIMIT 1",
					array(
						$row,
						$table,
						$idRow,
						$id
					), __FUNCTION__, __LINE__);

				// Row found?
				if (SQL_NUMROWS($result) == 1) {
					// Load the status
					list($currStatus) = SQL_FETCHROW($result);

					// And switch it N<->Y
					$newStatus = convertBooleanToYesNo(!($currStatus == 'Y'));

					// Change this status
					SQL_QUERY_ESC("UPDATE `{?_MYSQL_PREFIX?}_%s` SET %s='%s' WHERE %s=%s LIMIT 1",
						array(
							$table,
							$row,
							$newStatus,
							$idRow,
							$id
						), __FUNCTION__, __LINE__);

					// Count up affected rows
					$count += SQL_AFFECTEDROWS();
				} // END - if

				// Free the result
				SQL_FREERESULT($result);
			} // END - if
		} // END - foreach

		// Output status
		displayMessage(sprintf(getMessage('ADMIN_STATUS_CHANGED'), $count, count($IDs)));
	} else {
		// Nothing selected!
		displayMessage('{--ADMIN_NOTHING_SELECTED_CHANGE--}');
	}
}

// Send mails for del/edit/lock build modes
function sendAdminBuildMails ($mode, $table, $content, $id, $subjectPart = '', $userIdColumn = 'userid') {
	// Default subject is the subject part
	$subject = $subjectPart;

	// Is the subject part not set?
	if (empty($subjectPart)) {
		// Then use it from the mode
		$subject = strtoupper($mode);
	} // END - if

	// Is the raw userid set?
	if (postRequestParameter($userIdColumn, $id) > 0) {
		// Load email template
		if (!empty($subjectPart)) {
			$mail = loadEmailTemplate('member_' . $mode . '_' . strtolower($subjectPart) . '_' . $table, $content);
		} else {
			$mail = loadEmailTemplate('member_' . $mode . '_' . $table, $content);
		}

		// Send email out
		sendEmail(postRequestParameter($userIdColumn, $id), strtoupper('{--MEMBER_' . $subject . '_' . $table . '_SUBJECT--}'), $mail);
	} // END - if

	// Generate subject
	$subject = strtoupper('{--ADMIN_' . $subject . '_' . $table . '_SUBJECT--}');

	// Send admin notification out
	if (!empty($subjectPart)) {
		sendAdminNotification($subject, 'admin_' . $mode . '_' . strtolower($subjectPart) . '_' . $table, $content, postRequestParameter($userIdColumn, $id));
	} else {
		sendAdminNotification($subject, 'admin_' . $mode . '_' . $table, $content, postRequestParameter($userIdColumn, $id));
	}
}

// Build a special template list
function adminListBuilder ($listType, $IDs, $table, $columns, $filterFunctions, $extraValues, $idColumn, $userIdColumn, $userid = 'userid') {
	// $table and $idColumn must bove be arrays!
	if (!is_array($table)) {
		// $table is no array
		debug_report_bug(__FUNCTION__, __LINE__, 'table[]=' . gettype($table) . '!=array');
	} elseif (!is_array($idColumn)) {
		// $idColumn is no array
		debug_report_bug(__FUNCTION__, __LINE__, 'idColumn[]=' . gettype($idColumn) . '!=array');
	}

	$OUT = '';

	// "Walk" through all entries
	foreach ($IDs as $id => $selected) {
		// Secure id number
		$id = bigintval($id);

		// Get result from a given column array and table name
		$result = SQL_RESULT_FROM_ARRAY($table[0], $columns, $idColumn[0], $id, __FUNCTION__, __LINE__);

		// Is there one entry?
		if (SQL_NUMROWS($result) == 1) {
			// Load all data
			$content = SQL_FETCHARRAY($result);

			// Filter all data
			foreach ($content as $key => $value) {
				// Search index
				$idx = array_search($key, $columns, true);

				// Do we have a userid?
				if ($key == $userIdColumn) {
					// Add it again as raw id
					$content[$userIdColumn] = bigintval($value);
					$content[$userIdColumn . '_raw'] = $content[$userIdColumn];
				} // END - if

				// If the key matches the idColumn variable, we need to temporary remember it
				//* DEBUG: */ debugOutput('key=' . $key . ',idColumn=' . $idColumn . ',value=' . $value);
				if ($key == $idColumn) {
					// Found, so remember it
					$GLOBALS['admin_list_builder_id_value'] = $value;
				} // END - if

				// Handle the call in external function
				//* DEBUG: */ debugOutput('key=' . $key . ',fucntion=' . $filterFunctions[$idx] . ',value=' . $value);
				$content[$key] = handleExtraValues(
					$filterFunctions[$idx],
					$value,
					$extraValues[$idx]
				);
			} // END - foreach

			// Then list it
			$OUT .= loadTemplate(sprintf("admin_%s_%s_row",
				$listType,
				$table[0]
				), true, $content
			);
		} // END - if

		// Free the result
		SQL_FREERESULT($result);
	} // END - foreach

	// Load master template
	loadTemplate(sprintf("admin_%s_%s",
		$listType,
		$table[0]
		), false, $OUT
	);
}

// Change status of "build" list
function adminBuilderStatusHandler ($mode, $IDs, $table, $columns, $filterFunctions, $extraValues, $idColumn, $userIdColumn, $statusArray, $userid = 'userid') {
	// All valid entries? (We hope so here!)
	if ((is_array($IDs)) && (count($IDs) > 0) && (count($columns) == count($filterFunctions)) && (count($columns) == count($extraValues)) && (count($statusArray) > 0)) {
		// "Walk" through all entries
		foreach ($IDs as $id => $sel) {
			// Construct SQL query
			$sql = sprintf("UPDATE `{?_MYSQL_PREFIX?}_%s` SET", SQL_ESCAPE($table));

			// Load data of entry
			$result = SQL_QUERY_ESC("SELECT * FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`=%s LIMIT 1",
				array($table, $idColumn, $id), __FUNCTION__, __LINE__);

			// Fetch the data
			$content = SQL_FETCHARRAY($result);

			// Free the result
			SQL_FREERESULT($result);

			// Add all status entries (e.g. status column last_updated or so)
			$newStatus = 'UNKNOWN';
			$oldStatus = 'UNKNOWN';
			$statusColumn = 'unknown';
			foreach ($statusArray as $column => $statusInfo) {
				// Does the entry exist?
				if ((isset($content[$column])) && (isset($statusInfo[$content[$column]]))) {
					// Add these entries for update
					$sql .= sprintf(" `%s`='%s',", SQL_ESCAPE($column), SQL_ESCAPE($statusInfo[$content[$column]]));

					// Remember status
					if ($statusColumn == 'unknown') {
						// Always (!!!) change status column first!
						$oldStatus = $content[$column];
						$newStatus = $statusInfo[$oldStatus];
						$statusColumn = $column;
					} // END - if
				} elseif (isset($content[$column])) {
					// Unfinished!
					debug_report_bug(__FUNCTION__, __LINE__, ':UNFINISHED: id=' . $id . ',column=' . $column . '[' . gettype($statusInfo) . '] = ' . $content[$column]);
				}
			} // END - foreach

			// Add other columns as well
			foreach (postRequestArray() as $key => $entries) {
				// Debug message
				logDebugMessage(__FUNCTION__, __LINE__, 'Found entry: ' . $key);

				// Skip id, raw userid and 'do_$mode'
				if (!in_array($key, array($idColumn, $userid, ('do_' . $mode)))) {
					// Are there brackets () at the end?
					if (substr($entries[$id], -2, 2) == '()') {
						// Direct SQL command found
						$sql .= sprintf(" `%s`=%s,", SQL_ESCAPE($key), SQL_ESCAPE($entries[$id]));
					} else {
						// Add regular entry
						$sql .= sprintf(" `%s`='%s',", SQL_ESCAPE($key), SQL_ESCAPE($entries[$id]));

						// Add entry
						$content[$key] = $entries[$id];
					}
				} else {
					// Skipped entry
					logDebugMessage(__FUNCTION__, __LINE__, 'Skipped: ' . $key);
				}
			} // END - foreach

			// Finish SQL statement
			$sql = substr($sql, 0, -1) . sprintf(" WHERE `%s`=%s AND `%s`='%s' LIMIT 1",
				$idColumn,
				bigintval($id),
				$statusColumn,
				$oldStatus
			);

			// Run the SQL
			SQL_QUERY($sql, __FUNCTION__, __LINE__);

			// Do we have an URL?
			if (isset($content['url'])) {
				// Then add a framekiller test as well
				$content['frametester'] = generateFrametesterUrl($content['url']);
			} // END - if

			// Send "build mails" out
			sendAdminBuildMails($mode, $table, $content, $id, $statusInfo[$content[$column]], $userIdColumn);
		} // END - foreach
	} // END - if
}

// Delete rows by given id numbers
function adminDeleteEntriesConfirm ($IDs, $table, $columns = array(), $filterFunctions = array(), $extraValues = array(), $deleteNow = false, $idColumn = 'id', $userIdColumn = 'userid', $userid = 'userid') {
	// All valid entries? (We hope so here!)
	if ((is_array($IDs)) && (count($IDs) > 0) && (count($columns) == count($filterFunctions)) && (count($columns) == count($extraValues))) {
		// Shall we delete here or list for deletion?
		if ($deleteNow === true) {
			// The base SQL command:
			$sql = "DELETE LOW_PRIORITY FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s` IN (%s)";

			// Delete them all
			$idList = '';
			foreach ($IDs as $id => $sel) {
				// Is there a userid?
				if (isPostRequestParameterSet($userid, $id)) {
					// Load all data from that id
					$result = SQL_QUERY_ESC("SELECT * FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`=%s LIMIT 1",
						array(
							$table,
							$idColumn,
							$id
						), __FUNCTION__, __LINE__);

					// Fetch the data
					$content = SQL_FETCHARRAY($result);

					// Free the result
					SQL_FREERESULT($result);

					// Send "build mails" out
					sendAdminBuildMails('delete', $table, $content, $id, '', $userIdColumn);
				} // END - if

				// Add id number
				$idList .= $id . ',';
			} // END - foreach

			// Run the query
			SQL_QUERY_ESC($sql, array($table, $idColumn, substr($idList, 0, -1)), __FUNCTION__, __LINE__);

			// Was this fine?
			if (SQL_AFFECTEDROWS() == count($IDs)) {
				// All deleted
				displayMessage('{--ADMIN_ALL_ENTRIES_REMOVED--}');
			} else {
				// Some are still there :(
				displayMessage(sprintf(getMessage('ADMIN_SOME_ENTRIES_NOT_DELETED'), SQL_AFFECTEDROWS(), count($IDs)));
			}
		} else {
			// List for deletion confirmation
			adminListBuilder('delete', $IDs, $table, $columns, $filterFunctions, $extraValues, $idColumn, $userIdColumn);
		}
	} // END - if
}

// Edit rows by given id numbers
function adminEditEntriesConfirm ($IDs, $table, $columns = array(), $filterFunctions = array(), $extraValues = array(), $editNow = false, $idColumn = 'id', $userIdColumn = 'userid', $userid = 'userid') {
	// All valid entries? (We hope so here!)
	if ((is_array($IDs)) && (count($IDs) > 0) && (count($columns) == count($filterFunctions)) && (count($columns) == count($extraValues))) {
		// Shall we change here or list for editing?
		if ($editNow === true) {
			// Change them all
			$affected = '0';
			foreach ($IDs as $id => $sel) {
				// Prepare content array (new values)
				$content = array();

				// Prepare SQL for this row
				$sql = sprintf("UPDATE `{?_MYSQL_PREFIX?}_%s` SET",
					SQL_ESCAPE($table)
				);
				foreach (postRequestArray() as $key => $entries) {
					// Skip raw userid which is always invalid
					if ($key == $userid) {
						// Continue with next field
						continue;
					} // END - if

					// Is entries an array?
					if (($key != $idColumn) && (is_array($entries)) && (isset($entries[$id]))) {
						// Add this entry to content
						$content[$key] = $entries[$id];

						// Send data through the filter function if found
						if ((isset($filterFunctions[$key])) && (isset($extraValues[$key]))) {
							// Filter function set!
							$entries[$id] = handleExtraValues($filterFunctions[$key], $entries[$id], $extraValues[$key]);
						} // END - if

						// Then add this value
						$sql .= sprintf(" `%s`='%s',",
							SQL_ESCAPE($key),
							SQL_ESCAPE($entries[$id])
						);
					} elseif (($key != $idColumn) && (!is_array($entries))) {
						// Add normal entries as well!
						$content[$key] =  $entries;
					}

					// Do we have an URL?
					if ($key == 'url') {
						// Then add a framekiller test as well
						$content['frametester'] = generateFrametesterUrl($content[$key]);
					} // END - if
				} // END - foreach

				// Finish SQL command
				$sql = substr($sql, 0, -1) . " WHERE `" . $idColumn . "`=" . bigintval($id) . " LIMIT 1";

				// Run this query
				SQL_QUERY($sql, __FUNCTION__, __LINE__);

				// Add affected rows
				$affected += SQL_AFFECTEDROWS();

				// Load all data from that id
				$result = SQL_QUERY_ESC("SELECT * FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`=%s LIMIT 1",
					array($table, $idColumn, $id), __FUNCTION__, __LINE__);

				// Fetch the data and merge it into $content
				$content = merge_array($content, SQL_FETCHARRAY($result));

				// Free the result
				SQL_FREERESULT($result);

				// Send "build mails" out
				sendAdminBuildMails('edit', $table, $content, $id, '', $userIdColumn);
			} // END - foreach

			// Was this fine?
			if ($affected == count($IDs)) {
				// All deleted
				displayMessage('{--ADMIN_ALL_ENTRIES_EDITED--}');
			} else {
				// Some are still there :(
				displayMessage(sprintf(getMessage('ADMIN_SOME_ENTRIES_NOT_EDITED'), $affected, count($IDs)));
			}
		} else {
			// List for editing
			adminListBuilder('edit', $IDs, $table, $columns, $filterFunctions, $extraValues, $idColumn, $userIdColumn);
		}
	} else {
		// Maybe some invalid parameters
		debug_report_bug(__FUNCTION__, __LINE__, 'IDs[]=' . gettype($IDs) . ',table=' . $table . ',columns[]=' . gettype($columns) . ',filterFunctions[]=' . gettype($filterFunctions) . ',extraValues[]=' . gettype($extraValues) . ',idColumn=' . $idColumn . ',userIdColumn=' . $userIdColumn . ' - INVALID!');
	}
}

// Un-/lock rows by given id numbers
function adminLockEntriesConfirm ($IDs, $table, $columns=array(), $filterFunctions=array(), $extraValues=array(), $statusArray=array(), $lockNow=false, $idColumn='id', $userIdColumn='userid') {
	// All valid entries? (We hope so here!)
	if ((is_array($IDs)) && (count($IDs) > 0) && (count($columns) == count($filterFunctions)) && (count($columns) == count($extraValues)) && (($lockNow === false) || (count($statusArray) == 1))) {
		// Shall we un-/lock here or list for locking?
		if ($lockNow === true) {
			// Un-/lock entries
			adminBuilderStatusHandler('lock', $IDs, $table, $columns, $filterFunctions, $extraValues, $idColumn, $userIdColumn, $statusArray);
		} else {
			// List for editing
			adminListBuilder('lock', $IDs, $table, $columns, $filterFunctions, $extraValues, $idColumn, $userIdColumn);
		}
	} // END - if
}

// Undelete rows by given id numbers
function adminUndeleteEntriesConfirm ($IDs, $table, $columns=array(), $filterFunctions=array(), $extraValues=array(), $statusArray=array(), $undeleteNow=false, $idColumn='id', $userIdColumn='userid') {
	// All valid entries? (We hope so here!)
	if ((is_array($IDs)) && (count($IDs) > 0) && (count($columns) == count($filterFunctions)) && (count($columns) == count($extraValues)) && (($undeleteNow === false) || (count($statusArray) == 1))) {
		// Shall we un-/lock here or list for locking?
		if ($undeleteNow === true) {
			// Undelete entries
			adminBuilderStatusHandler('undelete', $IDs, $table, $columns, $filterFunctions, $extraValues, $idColumn, $userIdColumn, $statusArray);
		} else {
			// List for editing
			adminListBuilder('undelete', $IDs, $table, $columns, $filterFunctions, $extraValues, $idColumn, $userIdColumn);
		}
	} // END - if
}

// List all given rows (callback function from XML)
function adminListEntries ($tableTemplate, $rowTemplate, $noEntryMessageId, $tableName, $columns, $whereColumns, $orderByColumns, $callbackColumns, $extraParameters = array()) {
	// Verify that tableName and columns are not empty
	if (count($tableName) != 1) {
		// No tableName specified
		debug_report_bug(__FUNCTION__, __LINE__, 'tableName is not given. Please fix your XML. tableTemplate=' . $tableTemplate . ',rowTemplate=' . $rowTemplate);
	} elseif (count($columns) == 0) {
		// No columns specified
		debug_report_bug(__FUNCTION__, __LINE__, 'columns is not given. Please fix your XML. tableTemplate=' . $tableTemplate . ',rowTemplate=' . $rowTemplate . ',tableName[0]=' . $tableName[0]);
	}

	// This is the minimum query, so at least columns and tableName must have entries
	$SQL = 'SELECT ';
	foreach ($columns as $columnArray) {
		// Init SQL part
		$sqlPart = '';
		// Do we have a table/alias
		if (!empty($columnArray['table'])) {
			// Pre-add it
			$sqlPart .= $columnArray['table'] . '.';
		} // END - if

		// Add column
		$sqlPart .= '`' . $columnArray['column'] . '`';

		// Is a function and alias set?
		if ((!empty($columnArray['function'])) && (!empty($columnArray['alias']))) {
			// Add both
			$sqlPart = $columnArray['function'] . '(' . $sqlPart . ') AS `' . $columnArray['alias'] . '`';
		} // END - if

		// Add finished SQL part to the query
		$SQL .= $sqlPart . ',';
	} // END - foreach

	// Remove last commata and add FROM statement
	$SQL = substr($SQL, 0, -1) . ' FROM `{?_MYSQL_PREFIX?}_' . $tableName[0] . '`';

	// Do we have entries from whereColumns to add?
	if (count($whereColumns) > 0) {
		// Then add these as well
		if (count($whereColumns) == 1) {
			// One entry found
			$SQL .= ' WHERE ';

			// Table/alias included?
			if (!empty($whereColumns[0]['table'])) {
				// Add it as well
				$SQL .= $whereColumns[0]['table'] . '.';
			} // END - if

			// Add the rest
			$SQL .= '`' . $whereColumns[0]['column'] . '`' . $whereColumns[0]['condition'] . "'" . $whereColumns[0]['look_for'] . "'";
		} else {
			// More than one entry -> Unsupported
			debug_report_bug(__FUNCTION__, __LINE__, 'More than one WHERE statement found. This is currently not supported.');
		}
	} // END - if

	// Do we have entries from orderByColumns to add?
	if (count($orderByColumns) > 0) {
		// Add them as well
		$SQL .= ' ORDER BY ';
		foreach ($orderByColumns as $orderByColumn=>$array) {
			// Get keys (table/alias) and values (sorting itself)
			$table   = trim(implode('', array_keys($array)));
			$sorting = trim(implode('', array_keys($array)));

			// table/alias can be omitted
			if (!empty($table)) {
				// table/alias is given
				$SQL .= $table . '.';
			} // END - if

			// Add order-by column
			$SQL .= '`' . $orderByColumn . '` ' . $sorting . ',';
		} // END - foreach

		// Remove last column
		$SQL = substr($SQL, 0, -1);
	} // END - if

	// Now handle all over to the inner function which will execute the listing
	doAdminListEntries($SQL, $tableTemplate, $noEntryMessageId, $rowTemplate, $callbackColumns, $extraParameters = array());
}

// Do the listing of entries
function doAdminListEntries ($SQL, $tableTemplate, $noEntryMessageId, $rowTemplate, $callbackColumns, $extraParameters = array()) {
	// Run the SQL query
	$result = SQL_QUERY($SQL, __FUNCTION__, __LINE__);

	// Do we have some URLs left?
	if (!SQL_HASZERONUMS($result)) {
		// List all URLs
		$OUT = '';
		while ($content = SQL_FETCHARRAY($result)) {
			// "Translate" content
			foreach ($callbackColumns as $column=>$callbackFunction) {
				// Fill the callback arguments
				$args = array($content[$column]);

				// Do we have more to add?
				if (isset($extraParameters[$column])) {
					// Add them as well
					merge_array($args, $extraParameters[$column]);
				} // END - if

				// Call the callback-function
				//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'callbackFunction=' . $callbackFunction . ',args=<pre>'.print_r($args, true).'</pre>');
				// @TODO If we can rewrite the EL sub-system to support more than one parameter, this call_user_func_array() can be avoided
				$content[$column] = call_user_func_array($callbackFunction, $args);
			} // END - foreach

			// Load row template
			$OUT .= loadTemplate(trim($rowTemplate[0]), true, $content);
		} // END - while

		// Load main template
		loadTemplate(trim($tableTemplate[0]), false, $OUT);
	} else {
		// No URLs in surfbar
		displayMessage('{--' .$noEntryMessageId . '--}');
	}

	// Free result
	SQL_FREERESULT($result);
}

// Checks proxy settins by fetching check-updates3.php from www.mxchange.org
function adminTestProxySettings ($settingsArray) {
	// Set temporary the new settings
	mergeConfig($settingsArray);

	// Now get the test URL
	$content = sendGetRequest('check-updates3.php');

	// Is the first line with "200 OK"?
	$valid = (strpos($content[0], '200 OK') !== false);

	// Return result
	return $valid;
}

// Sends out a link to the given email adress so the admin can reset his/her password
function sendAdminPasswordResetLink ($email) {
	// Init output
	$OUT = '';

	// Look up administator login
	$result = SQL_QUERY_ESC("SELECT `id`, `login`, `password` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `email`='%s' LIMIT 1",
		array($email), __FUNCTION__, __LINE__);

	// Is there an account?
	if (SQL_HASZERONUMS($result)) {
		// No account found
		return '{--ADMIN_NO_LOGIN_WITH_EMAIL--}';
	} // END - if

	// Load all data
	$content = SQL_FETCHARRAY($result);

	// Free result
	SQL_FREERESULT($result);

	// Generate hash for reset link
	$content['hash'] = generateHash(getUrl() . getEncryptSeperator() . $content['id'] . getEncryptSeperator() . $content['login'] . getEncryptSeperator() . $content['password'], substr($content['password'], getSaltLength()));

	// Remove some data
	unset($content['id']);
	unset($content['password']);

	// Prepare email
	$mailText = loadEmailTemplate('admin_reset_password', $content);

	// Send it out
	sendEmail($email, '{--ADMIN_RESET_PASSWORD_LINK_SUBJECT--}', $mailText);

	// Prepare output
	return '{--ADMIN_RESET_PASSWORD_LINK_SENT--}';
}

// Validate hash and login for password reset
function adminResetValidateHashLogin ($hash, $login) {
	// By default nothing validates... ;)
	$valid = false;

	// Then try to find that user
	$result = SQL_QUERY_ESC("SELECT `id`, `password`, `email` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `login`='%s' LIMIT 1",
		array($login), __FUNCTION__, __LINE__);

	// Is an account here?
	if (SQL_NUMROWS($result) == 1) {
		// Load all data
		$content = SQL_FETCHARRAY($result);

		// Generate hash again
		$hashFromData = generateHash(getUrl() . getEncryptSeperator() . $content['id'] . getEncryptSeperator() . $login . getEncryptSeperator() . $content['password'], substr($content['password'], getSaltLength()));

		// Does both match?
		$valid = ($hash == $hashFromData);
	} // END - if

	// Free result
	SQL_FREERESULT($result);

	// Return result
	return $valid;
}

// Reset the password for the login. Do NOT call this function without calling above function first!
function doResetAdminPassword ($login, $password) {
	// Generate hash (we already check for sql_patches in generateHash())
	$passHash = generateHash($password);

	// Prepare fake POST data
	$postData = array(
		'login'    => array(getAdminId($login) => $login),
		'password' => array(getAdminId($login) => $passHash),
	);

	// Update database
	$message = adminsChangeAdminAccount($postData, '', false);

	// Run filters
	runFilterChain('post_form_reset_pass', array('login' => $login, 'hash' => $passHash, 'message' => $message));

	// Return output
	return '{--ADMIN_PASSWORD_RESET_DONE--}';
}

// Solves a task by given id number
function adminSolveTask ($id) {
	// Update the task data
	adminUpdateTaskData($id, 'status', 'SOLVED');
}

// Marks a given task as deleted
function adminDeleteTask ($id) {
	// Update the task data
	adminUpdateTaskData($id, 'status', 'DELETED');
}

// Function to update task data
function adminUpdateTaskData ($id, $row, $data) {
	// Should be admin!
	if (!isAdmin()) {
		// Not an admin so redirect better
		redirectToUrl('modules.php?module=index');
	} // END - if

	// Is the id not set, then we need a backtrace here... :(
	if ($id <= 0) {
		// Initiate backtrace
		debug_report_bug(__FUNCTION__, __LINE__, sprintf("id is invalid: %s. row=%s, data=%s",
			$id,
			$row,
			$data
		));
	} // END - if

	// Update the task
	SQL_QUERY_ESC("UPDATE `{?_MYSQL_PREFIX?}_task_system` SET `%s`='%s' WHERE `id`=%s LIMIT 1",
		array(
			$row,
			$data,
			bigintval($id)
		), __FUNCTION__, __LINE__);
}

// Checks wether if the admin menu has entries
function ifAdminMenuHasEntries ($action) {
	return (
		((
			// Is the entry set?
			isset($GLOBALS['admin_menu_has_entries'][$action])
		) && (
			// And do we have a menu for this action?
			$GLOBALS['admin_menu_has_entries'][$action] === true
		)) || (
			// Login has always a menu
			$action == 'login'
		)
	);
}

// Setter for 'admin_menu_has_entries'
function setAdminMenuHasEntries ($action, $hasEntries) {
	$GLOBALS['admin_menu_has_entries'][$action] = (bool) $hasEntries;
}

// Creates a link to the user's admin-profile
function adminCreateUserLink ($userid) {
	// Is the userid set correctly?
	if (isValidUserId($userid)) {
		// Create a link to that profile
		return '{%url=modules.php?module=admin&amp;what=list_user&amp;userid=' . bigintval($userid) . '%}';
	} // END - if

	// Return a link to the user list
	return '{%url=modules.php?module=admin&amp;what=list_user%}';
}

// Generate a "link" for the given admin id (admin_id)
function generateAdminLink ($adminId) {
	// No assigned admin is default
	$adminLink = '<span class="notice">{--ADMIN_NO_ADMIN_ASSIGNED--}</span>';

	// Zero? = Not assigned
	if (bigintval($adminId) > 0) {
		// Load admin's login
		$login = getAdminLogin($adminId);

		// Is the login valid?
		if ($login != '***') {
			// Is the extension there?
			if (isExtensionActive('admins')) {
				// Admin found
				$adminLink = '<a href="' . generateEmailLink(getAdminEmail($adminId), 'admins') . '" title="{--ADMIN_CONTACT_LINK_TITLE--}">' . $login . '</a>';
			} else {
				// Extension not found
				$adminLink = '{%message,ADMIN_TASK_ROW_EXTENSION_NOT_INSTALLED=admins%}';
			}
		} else {
			// Maybe deleted?
			$adminLink = '<div class="notice">{%message,ADMIN_ID_404=' . $adminId . '%}</div>';
		}
	} // END - if

	// Return result
	return $adminLink;
}

// Verifies if the current admin has confirmed to alter expert settings
//
// Return values:
// 'failed'    = Something goes wrong (default)
// 'agreed'    = Has verified and and confirmed it to see them
// 'forbidden' = Has not the proper right to alter them
// 'update'    = Need to update extension 'admins'
// 'ask'       = A form was send to the admin
function doVerifyExpertSettings () {
	// Default return status is failed
	$return = 'failed';

	// Is the extension installed and recent?
	if (isExtensionInstalledAndNewer('admins', '0.7.3')) {
		// Okay, load the status
		$expertSettings = getAminsExpertSettings();

		// Is he allowed?
		if ($expertSettings == 'Y') {
			// Okay, does he want to see them?
			if (isAdminsExpertWarningEnabled()) {
				// Ask for them
				if (isFormSent()) {
					// Is the element set, then we need to change the admin
					if (isPostRequestParameterSet('expert_settings')) {
						// Get it and prepare final post data array
						$postData['login'][getCurrentAdminId()] = getCurrentAdminLogin();
						$postData['expert_warning'][getCurrentAdminId()] = 'N';

						// Change it in the admin
						adminsChangeAdminAccount($postData, 'expert_warning');

						// Clear form
						unsetPostRequestParameter('ok');
					} // END - if

					// All fine!
					$return = 'agreed';
				} else {
					// Send form
					loadTemplate('admin_expert_settings_form');

					// Asked for it
					$return = 'ask';
				}
			} else {
				// Do not display
				$return = 'agreed';
			}
		} else {
			// Forbidden
			$return = 'forbidden';
		}
	} else {
		// Out-dated extension or not installed
		$return = 'update';
	}

	// Output message for other status than ask/agreed
	if (($return != 'ask') && ($return != 'agreed')) {
		// Output message
		displayMessage('{--ADMIN_EXPERT_SETTINGS_STATUS_' . strtoupper($return) . '--}');
	} // END - if

	// Return status
	return $return;
}

// Generate link to unconfirmed mails for admin
function generateUnconfirmedAdminLink ($id, $unconfirmed, $type = 'bid') {
	// Init output
	$OUT = $unconfirmed;

	// Do we have unconfirmed mails?
	if ($unconfirmed > 0) {
		// Add link to list_unconfirmed what-file
		$OUT = '<a href="{%url=modules.php?module=admin&amp;what=list_unconfirmed&amp;' . $type . '=' . $id . '%}">{%pipe,translateComma=' . $unconfirmed . '%}</a>';
	} // END - if

	// Return it
	return $OUT;
}

// Generates a navigation row for listing emails
function addEmailNavigation ($numPages, $offset, $show_form, $colspan, $return=false) {
	// Don't do anything if $numPages is 1
	if ($numPages == 1) {
		// Abort here with empty content
		return '';
	} // END - if

	$TOP = '';
	if ($show_form === false) {
		$TOP = ' top';
	} // END - if

	$NAV = '';
	for ($page = 1; $page <= $numPages; $page++) {
		// Is the page currently selected or shall we generate a link to it?
		if (($page == getRequestParameter('page')) || ((!isGetRequestParameterSet('page')) && ($page == 1))) {
			// Is currently selected, so only highlight it
			$NAV .= '<strong>-';
		} else {
			// Open anchor tag and add base URL
			$NAV .= '<a href="{%url=modules.php?module=admin&amp;what=' . getWhat() . '&amp;page=' . $page . '&amp;offset=' . $offset;

			// Add userid when we shall show all mails from a single member
			if ((isGetRequestParameterSet('userid')) && (isValidUserId(getRequestParameter('userid')))) $NAV .= '&amp;userid=' . bigintval(getRequestParameter('userid'));

			// Close open anchor tag
			$NAV .= '%}">';
		}
		$NAV .= $page;
		if (($page == getRequestParameter('page')) || ((!isGetRequestParameterSet('page')) && ($page == 1))) {
			// Is currently selected, so only highlight it
			$NAV .= '-</strong>';
		} else {
			// Close anchor tag
			$NAV .= '</a>';
		}

		// Add seperator if we have not yet reached total pages
		if ($page < $numPages) {
			// Add it
			$NAV .= '|';
		} // END - if
	} // END - for

	// Define constants only once
	$content['nav']  = $NAV;
	$content['span'] = $colspan;
	$content['top']  = $TOP;

	// Load navigation template
	$OUT = loadTemplate('admin_email_nav_row', true, $content);

	if ($return === true) {
		// Return generated HTML-Code
		return $OUT;
	} else {
		// Output HTML-Code
		outputHtml($OUT);
	}
}

// Process menu editing form
function adminProcessMenuEditForm ($type, $subMenu) {
	// An action is done...
	foreach (postRequestParameter('sel') as $sel => $menu) {
		$AND = "(`what` = '' OR `what` IS NULL)";

		$sel = bigintval($sel);

		if (!empty($subMenu)) {
			$AND = "`action`='" . $subMenu . "'";
		} // END - if

		switch (postRequestParameter('ok')) {
			case 'edit': // Edit menu
				if (postRequestParameter('sel_what', $sel) == '') {
					// Update with 'what'=null
					SQL_QUERY_ESC("UPDATE `{?_MYSQL_PREFIX?}_%s_menu` SET `title`='%s', `action`='%s', `what`=NULL WHERE ".$AND." AND `id`=%s LIMIT 1",
						array(
							$type,
							$menu,
							postRequestParameter('sel_action', $sel),
							$sel
						), __FILE__, __LINE__);
				} else {
					// Update with selected 'what'
					SQL_QUERY_ESC("UPDATE `{?_MYSQL_PREFIX?}_%s_menu` SET `title`='%s', `action`='%s', `what`='%s' WHERE ".$AND." AND `id`=%s LIMIT 1",
						array(
							$type,
							$menu,
							postRequestParameter('sel_action', $sel),
							postRequestParameter('sel_what', $sel),
							$sel
						), __FILE__, __LINE__);
				}
				break;

			case 'delete': // Delete menu
				SQL_QUERY_ESC("DELETE LOW_PRIORITY FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE ".$AND." AND `id`=%s LIMIT 1",
					array($type, $sel), __FILE__, __LINE__);
				break;

			case 'status': // Change status of menus
				SQL_QUERY_ESC("UPDATE `{?_MYSQL_PREFIX?}_%s_menu` SET `visible`='%s', `locked`='%s' WHERE ".$AND." AND `id`=%s LIMIT 1",
					array($type, postRequestParameter('visible', $sel), postRequestParameter('locked', $sel), $sel), __FILE__, __LINE__);
				break;

			default: // Unexpected action
				logDebugMessage(__FILE__, __LINE__, sprintf("Unsupported action %s detected.", postRequestParameter('ok')));
				displayMessage('{%message,ADMIN_UNKNOWN_OKAY=' . postRequestParameter('ok') . '%}');
				break;
		} // END - switch
	} // END - foreach

	// Load template
	displayMessage('{--SETTINGS_SAVED--}');
}

// Handle weightning
function doAdminProcessMenuWeightning ($type, $AND) {
	// Are there all required (generalized) GET parameter?
	if ((isGetRequestParameterSet('act')) && (isGetRequestParameterSet('tid')) && (isGetRequestParameterSet('fid'))) {
		// Init variables
		$tid = ''; $fid = '';

		// Get ids
		if (isGetRequestParameterSet('w')) {
			// Sub menus selected
			$result = SQL_QUERY_ESC("SELECT `id` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE `action`='%s' AND `sort`=%s LIMIT 1",
				array(
					$type,
					getRequestParameter('act'),
					bigintval(getRequestParameter('tid'))
				), __FILE__, __LINE__);
			list($tid) = SQL_FETCHROW($result);
			SQL_FREERESULT($result);
			$result = SQL_QUERY_ESC("SELECT `id` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE `action`='%s' AND `sort`=%s LIMIT 1",
				array(
					$type,
					getRequestParameter('act'),
					bigintval(getRequestParameter('fid'))
				), __FILE__, __LINE__);
			list($fid) = SQL_FETCHROW($result);
			SQL_FREERESULT($result);
		} else {
			// Main menu selected
			$result = SQL_QUERY_ESC("SELECT `id` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE (`what`='' OR `what` IS NULL) AND `sort`=%s LIMIT 1",
				array(
					$type,
					bigintval(getRequestParameter('tid'))
				), __FILE__, __LINE__);
			list($tid) = SQL_FETCHROW($result);
			SQL_FREERESULT($result);
			$result = SQL_QUERY_ESC("SELECT `id` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE (`what`='' OR `what` IS NULL) AND `sort`=%s LIMIT 1",
				array(
					$type,
					bigintval(getRequestParameter('fid'))
				), __FILE__, __LINE__);
			list($fid) = SQL_FETCHROW($result);
			SQL_FREERESULT($result);
		}

		if ((!empty($tid)) && (!empty($fid))) {
			// Sort menu
			SQL_QUERY_ESC("UPDATE `{?_MYSQL_PREFIX?}_%s_menu` SET `sort`=%s WHERE ".$AND." AND `id`=%s LIMIT 1",
				array(
					$type,
					bigintval(getRequestParameter('tid')),
					bigintval($fid)
				), __FILE__, __LINE__);
			SQL_QUERY_ESC("UPDATE `{?_MYSQL_PREFIX?}_%s_menu` SET `sort`=%s WHERE ".$AND." AND `id`=%s LIMIT 1",
				array(
					$type,
					bigintval(getRequestParameter('fid')),
					bigintval($tid)
				), __FILE__, __LINE__);
		} // END - if
	} // END - if
}

// [EOF]
?>
